@@ -1,3 +1,4 @@
+"""A biochemical reaction module."""
 # The MIT License (MIT)
 #
 # Copyright (c) 2018 Institute for Molecular Systems Biology, ETH Zurich.
@@ -26,7 +27,8 @@
 import logging
 import re
 import sys
-from typing import Callable, Dict, Iterable, List, Optional, Tuple
+import warnings
+from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
 
 import numpy as np
 import pandas as pd
@@ -34,7 +36,7 @@ import pandas as pd
 from . import PROTON_INCHI, WATER_INCHI
 from .exceptions import ParseException
 from .models import Compound
-from .thermodynamic_constants import Q_, ureg
+from .thermodynamic_constants import Q_, default_pMg, ureg
 
 
 __all__ = ("POSSIBLE_REACTION_ARROWS", "Reaction")
@@ -61,64 +63,112 @@ POSSIBLE_REACTION_ARROWS = (
 
 
 class Reaction:
+    """A class for handling biochemical reactions."""
+
     def __init__(
         self,
         sparse: Dict[Compound, float],
         arrow: Optional[str] = None,
         rid: Optional[str] = None,
     ):
+        """Create a new Reaction object."""
         self.sparse = {k: v for k, v in sparse.items() if v != 0}
         self.arrow = arrow or POSSIBLE_REACTION_ARROWS[0]
         self.rid = rid
 
     def __len__(self) -> int:
+        """Count the number of reactants."""
         return len(self.sparse)
 
     def clone(self) -> "Reaction":
+        """Create a copy of this reaction."""
         return Reaction(self.sparse, self.arrow, self.rid)
 
     def keys(self, protons=True, water=True) -> Iterable[Compound]:
-        gen = self.sparse.keys()
-        if water and protons:
-            return gen
-        elif water:
-            return (c for c in gen if c.inchi != PROTON_INCHI)
-        elif protons:
-            return (c for c in gen if c.inchi != WATER_INCHI)
-        else:
-            return (
-                c
-                for c in gen
-                if c.inchi != PROTON_INCHI and c.inchi != WATER_INCHI
-            )
+        """Iterate over the reactants.
 
-    def items(self, protons=True, water=True) -> Iterable[Tuple[Compound, int]]:
-        gen = self.sparse.items()
-        if water and protons:
-            return gen
-        elif water:
-            return ((c, i) for c, i in gen if c.inchi != PROTON_INCHI)
-        elif protons:
-            return ((c, i) for c, i in gen if c.inchi != WATER_INCHI)
-        else:
-            return (
-                (c, i)
-                for c, i in gen
-                if c.inchi != PROTON_INCHI and c.inchi != WATER_INCHI
-            )
+        Parameters
+        ----------
+        protons : bool
+             Include H+ as one of the keys (Default value = True)
+        water : bool
+             Include H2O as one of the keys (Default value = True)
+
+        Returns
+        -------
+        Iterable[Compound]
+            An iterator over all reactants.
+
+        """
+        for c in self.sparse.keys():
+            if not water and c.inchi == WATER_INCHI:
+                continue
+            if not protons and c.inchi == PROTON_INCHI:
+                continue
+            yield c
+
+    def items(
+        self, protons=True, water=True
+    ) -> Iterable[Tuple[Compound, float]]:
+        """Iterate over the reactants and their stoichiometric coefficients.
+
+        Parameters
+        ----------
+        protons : bool
+             Include H+ as one of the keys (Default value = True)
+        water : bool
+             Include H2O as one of the keys (Default value = True)
+
+        Returns
+        -------
+        Iterable[Tuple[Compound, int]]
+            An iterator of all reactants where the values are tuples of
+            Compound and stoichiometric coefficient
+
+        """
+        for c, i in self.sparse.items():
+            if not water and c.inchi == WATER_INCHI:
+                continue
+            if not protons and c.inchi == PROTON_INCHI:
+                continue
+            yield c, i
 
     def get_coeff(self, compound: Compound) -> float:
+        """Get the stoichiometric coefficient of a reactant.
+
+        Parameters
+        ----------
+        compound : Compound
+            A reactant Compound object
+
+        Returns
+        -------
+        float
+            The stoichiometric coefficient.
+
+
+        """
         return self.sparse.get(compound, 0.0)
 
     def reverse(self) -> "Reaction":
+        """Reverse the direction of the reaction.
+
+        The reverse reaction is calculated by negating all stoichiometric
+        coefficients.
+
+
+        Returns
+        -------
+        Reaction
+            The reverse reaction.
+
         """
-            reverse the direction of the reaction by negating all
-            stoichiometric coefficients
-        """
+
         sparse = dict((k, -v) for (k, v) in self.sparse.items())
         return Reaction(sparse, self.arrow, self.rid)
 
     def __eq__(self, other: "Reaction") -> bool:
+        """Check if two reactions are equal."""
         cpds = set(self.keys(protons=False)).union(other.keys(protons=False))
         for c in cpds:
             if self.get_coeff(c) != other.get_coeff(c):
@@ -129,13 +179,25 @@ class Reaction:
     def parse_formula_side(
         s: str, str_to_compound: Callable[[str], Compound]
     ) -> Dict[Compound, float]:
-        """
-            Parses the side formula, e.g.
-            '2 KEGG:C00001 + KEGG:C00002 + 3 KEGG:C00003'
-            Ignores stoichiometry.
+        """Parse one side of the reaction formula.
+
+        For example, '2 kegg:C00001 + kegg:C00002 + 3 kegg:C00003'.
+        Ignores stoichiometry.
+
+        Parameters
+        ----------
+        s: str :
+
+        str_to_compound: Callable[[str] :
+
+        Compound] :
+
+
+        Returns
+        -------
+        dict
+            The set of CIDs.
 
-            Returns:
-                The set of CIDs.
         """
         if s.strip() == "null":
             return {}
@@ -171,18 +233,24 @@ class Reaction:
         formula: str,
         rid: Optional[str] = None,
     ) -> "Reaction":
-        """
-        Parse a two-sided reaction definition.
+        """Parse a two-sided reaction formula.
+
+        Examples
+        --------
+        >>> from equilibrator_cache import Reaction
+        >>> Reaction.parse_formula(parse_compound, '2 C00001 = C00002 + C00003')
+
 
         Parameters
         ----------
         str_to_compound : Callable[[str], Compound]
-            A function that returns a compound instance from a string,
-            for example, from a database identifier.
+            A function that returns a compound instance
+            from a string, for example, from a database identifier.
         formula : str
             A string representation of the reaction.
         rid : str, optional
             An identifier for the resulting reaction object.
+            (Default value = None)
 
         Returns
         -------
@@ -190,11 +258,6 @@ class Reaction:
             An object representation of the substrates, products and the
             reaction direction.
 
-        Examples
-        --------
-        >>> from  equilibrator_cache import Reaction
-        >>> Reaction.parse_formula(parse_compound, '2 C00001 = C00002 + C00003')
-
         """
         tokens = []
         arrow = None
@@ -230,13 +293,31 @@ class Reaction:
 
     @staticmethod
     def write_compound_and_coeff(compound: Compound, coeff: float) -> str:
+        """Write the compound preceded by the stoichiometric coefficient.
+
+        Parameters
+        ----------
+        compound : Compound
+            A Compound object
+
+        coeff : float
+            The stoichiometric coefficient
+
+
+        Returns
+        -------
+        str
+            A string representation of the compound preceded by the
+            stoichiometric coefficient.
+
+        """
         if np.abs(coeff - 1) < sys.float_info.epsilon:
             return str(compound)
         else:
             return "%g %s" % (coeff, str(compound))
 
     def __str__(self) -> str:
-        """String representation."""
+        """Return a string representation of the entire reaction."""
         left = []
         right = []
         for compound, coeff in sorted(self.sparse.items()):
@@ -247,15 +328,13 @@ class Reaction:
         return f"{' + '.join(left)} {self.arrow} {' + '.join(right)}"
 
     def get_element_data_frame(self) -> pd.DataFrame:
-        """
-        Tabulate the elemental composition of all reactants.
+        """Tabulate the elemental composition of all reactants.
 
         Returns
         -------
-        pandas.DataFrame
+        DataFrame
             A data frame where the columns are the compounds and the
             indexes are atomic elements.
-
         """
         atom_bags = {
             compound: compound.atom_bag or {} for compound in self.keys()
@@ -268,19 +347,25 @@ class Reaction:
         ).fillna(0)
 
     def _get_reaction_atom_bag(
-        self, raise_exception: bool = False, minimal_stoichiometry: float = None
+        self, minimal_stoichiometry: Optional[float] = None
     ) -> Dict[str, int]:
-        """
-            Use for checking if all elements are conserved.
+        """Use for checking if all elements are conserved.
 
-            :param raise_exception: if True, raises exception if any of the
-            compounds does not have a proper formula
-            :param minimal_stoichiometry: if not None, sets the minimal value
+        Parameters
+        ----------
+        minimal_stoichiometry : float, optional
+            if not None, sets the minimal value
             for a non-zero stoichiometry
-            Returns:
-                An atom_bag of the differences between the sides of the
-                reaction. E.g. if there is one extra C on the left-hand
-                side, the result will be {'C': -1}.
+            (Default value = None)
+
+
+        Returns
+        -------
+        dict
+            An atom_bag of the differences between the sides of the
+            reaction. E.g. if there is one extra C on the left-hand
+            side, the result will be {'C': -1}.
+
         """
         element_df = self.get_element_data_frame()
 
@@ -290,11 +375,8 @@ class Reaction:
                 + "compounds have unspecific formulas: "
                 + "%s" % str(self)
             )
-            if raise_exception:
-                raise ValueError(warning_str)
-            else:
-                logging.warning(warning_str)
-                return None
+            warnings.warn(warning_str)
+            return None
 
         stoichiometry = np.array(list(element_df.columns.map(self.get_coeff)))
         unbalanced = element_df @ stoichiometry
@@ -323,6 +405,15 @@ class Reaction:
         """Add to an existing stoichiometric coefficient.
 
         If the compound is not part of the reaction, add it as a new reactant.
+
+        Parameters
+        ----------
+        cpd : Compound
+            the reactant
+        coeff : float
+            how much to add to its stoichiometric coefficient
+
+
         """
         if cpd in self.sparse:
             if self.sparse[cpd] == -coeff:
@@ -332,25 +423,27 @@ class Reaction:
         else:
             self.sparse[cpd] = coeff
 
-    def is_balanced(
-        self, ignore_atoms: List[str] = None, raise_exception: bool = False
-    ) -> bool:
-        """Check if a reaction is balanced
+    def is_balanced(self, ignore_atoms: Tuple[str] = ("H",)) -> bool:
+        """Check if a reaction is balanced.
+
+        Parameters
+        ----------
+        ignore_atoms : tuple, optional
+            a tuple with all the elements to ignore
+            (Default value = ("H",) )
+
+        Returns
+        -------
+        bool
+            True if the reaction is balanced.
 
-        :param ignore_atoms: tuple containing atoms to ignore (default: only H)
-        :param raise_exception: raise Exception if formulas are missing
-        :return:
         """
-        reaction_atom_bag = self._get_reaction_atom_bag(
-            raise_exception=raise_exception
-        )
+        reaction_atom_bag = self._get_reaction_atom_bag()
 
         if reaction_atom_bag is None:
             # this means some compound formulas are missing
             return False
 
-        if ignore_atoms is None:
-            ignore_atoms = ["H"]
         for atom in ignore_atoms:
             if atom in reaction_atom_bag:
                 reaction_atom_bag.pop(atom)
@@ -358,110 +451,209 @@ class Reaction:
         return len(reaction_atom_bag) == 0
 
     def balance_with_compound(
-        self,
-        compound: Compound,
-        ignore_atoms: Tuple[str] = ("H",),
-        raise_exception: bool = False,
-    ):
-        """Try to balance the reaction using a specified compound
+        self, compound: Compound, ignore_atoms: Tuple[str] = ("H",)
+    ) -> Union["Reaction", None]:
+        """Try to balance the reaction using a specified compound.
+
+        Parameters
+        ----------
+        compound : Compound
+            the compound to use for balancing
+        ignore_atoms : Tuple[str]
+            a tuple with all the elements to ignore
+            (Default value = ("H",) )
+
+        Returns
+        -------
+        Union[Reaction, None]
+            A balanced reaction. If the original reaction is already
+            balanced, returns it back. Otherwise, return a new balanced
+            reaction, or None if the reaction cannot be balanced.
 
-        :param compound: the compound to use for balancing
-        :param ignore_atoms: a tuple with all the elements to ignore
-        :param raise_exception: raise Exception if formulas are missing
-        :return: A balanced reaction. If the original reaction is already
-        balanced, returns it back. Otherwise, return a new balanced reaction,
-        or None if the reaction cannot be balanced.
         """
 
-        def get_pivot_atom(compound: Compound) -> Tuple[str, float]:
-            # select one atom from the given compound and try to balance it
-            # (skip protons if we decided to ignore them)
-            if compound.atom_bag is None:
+        def get_pivot_atom(_compound: Compound) -> Tuple[str, float]:
+            """Select one atom from the given compound and try to balance it.
+
+            Skips the elements in `ignore_atoms`
+
+            Parameters
+            ----------
+            _compound : Compound
+                the compound to use for balancing
+
+
+            Returns
+            -------
+            atom : str
+                the symbol of the element used for balancing
+            count : float
+                the number of atoms of this type in `compound`
+            """
+
+            if _compound.atom_bag is None:
                 raise Exception(
                     f"Cannot balance using this compound, it has no "
-                    f"formula: {compound.formula}"
+                    f"formula: {_compound.formula}"
                 )
 
-            for atom, count in compound.atom_bag.items():
-                if atom not in ignore_atoms:
+            for atom, count in _compound.atom_bag.items():
+                if ignore_atoms is None or atom not in ignore_atoms:
                     return atom, count
             raise Exception(
                 f"Cannot balance using this compound, it has no "
-                f"relevant atoms: {compound.formula}"
+                f"relevant atoms: {_compound.formula}"
             )
 
-        if self.is_balanced(ignore_atoms, raise_exception):
+        if self.is_balanced(ignore_atoms):
             return self
 
-        reaction_atom_bag = self._get_reaction_atom_bag(
-            raise_exception=raise_exception
-        )
+        reaction_atom_bag = self._get_reaction_atom_bag()
 
         if reaction_atom_bag is None:
             # this means some compound formulas are missing
             return None
 
-        pivot_atom, count = get_pivot_atom(compound)
+        pivot_atom, pivot_count = get_pivot_atom(compound)
 
         new_reaction = self.clone()
         if pivot_atom in reaction_atom_bag:
             new_reaction.add_stoichiometry(
-                compound, -reaction_atom_bag[pivot_atom] / float(count)
+                compound, -reaction_atom_bag[pivot_atom] / float(pivot_count)
             )
 
-        if new_reaction.is_balanced(ignore_atoms, raise_exception):
+        if new_reaction.is_balanced(ignore_atoms):
             return new_reaction
 
-    def is_empty(self):
+    def is_empty(self) -> bool:
+        """Check if the reaction is empty.
+
+        Returns
+        -------
+        bool
+            whether the reaction is empty (i.e. has no reactants).
+
+        """
         return len(self.sparse) == 0
 
-    def dense(self, cids):
+    def dense(self, cids: List[Compound]) -> np.ndarray:
+        """Return a dense representation of the reaction.
+
+        Parameters
+        ----------
+        cids : List[Compound]
+            The order of compounds for determining the indices in the result
+
+
+        Returns
+        -------
+        np.ndarray
+            An array with the stoichiometric coefficients in "dense" format.
+
+        """
         s = np.zeros((len(cids), 1))
         for cid, coeff in self.items():
             s[cids.index(cid), 0] = coeff
         return s
 
-    @ureg.check(None, "", "[concentration]", "[temperature]")
-    def transform(self, p_h: Q_, ionic_strength: Q_, temperature: Q_) -> float:
-        """Use the Legendre transform to convert the ddG_over_RT to the
-        difference in the transformed energies of this MS and the major MS
+    def can_be_transformed(self) -> bool:
+        """Check if this reaction can be transformed.
+
+        In other words, all the reactants can be transformed (i.e. have
+        microspecies).
+
+        Returns
+        -------
+        True if the reaction can be transformed
 
-        :param p_h:
-        :param ionic_strength:
-        :param temperature:
-        :return: the transformed relative deltaG (in units of RT)
         """
-        ddg_over_rt = Q_(0.0)
-        # Protons are ignored in the Legendre transform.
-        for compound, coeff in self.items(protons=False):
-            ddg_over_rt += coeff * compound.transform(
-                p_h, ionic_strength, temperature
+        for compound, _ in self.items():
+            if not compound.can_be_transformed():
+                return False
+        return True
+
+    @ureg.check(None, "", "[concentration]", "[temperature]", "")
+    def transform(
+        self,
+        p_h: Q_,
+        ionic_strength: Q_,
+        temperature: Q_,
+        p_mg: Q_ = default_pMg,
+    ) -> Q_:
+        r"""Calculate the Legendre transform for a reaction.
+
+        Apply the Legendre transform on each of the compounds and sum up their
+        contributions (based on the stoichiometry).
+
+        Parameters
+        ----------
+        p_h : Quantity
+            Set the -log10 of the proton concentration [H+].
+        ionic_strength : Quantity
+            Set the ionic strength.
+        temperature : Quantity
+            Set the temperature.
+        p_mg : Quantity
+            Set the -log10 of the magnesium ion concentration [Mg++].
+            (Default value = default_pMg)
+
+        Returns
+        -------
+        Quantity
+            The transformed relative :math:`\Delta G`.
+
+        """
+        ddg = Q_(0.0, "kJ/mol")
+        for compound, coeff in self.items(protons=False):  # ignore protons
+            ddg += coeff * compound.transform(
+                p_h=p_h,
+                ionic_strength=ionic_strength,
+                temperature=temperature,
+                p_mg=p_mg,
             )
-        return ddg_over_rt
+        return ddg
 
     def _sum_coefficients(self) -> float:
-        """
-        Calculate the sum of all coefficients (excluding water).
+        r"""Calculate the sum of all coefficients (excluding water).
 
-        This is useful for shifting the dG'0 to another set of standard
-        concentrations (e.g. 1 mM)
+        This is useful for shifting the :math:`\Delta G0`to another set
+        of standard concentrations (e.g. 1 mM).
 
+
+        Returns
+        -------
+        float
+            The sum of all coefficients.
         """
+
         return sum(c for _, c in self.items(protons=False, water=False))
 
     def _sum_absolute_coefficients(self) -> float:
+        """Calculate sum of coefficients (excluding water) in absolute value.
+
+        This is useful for calculating the reversibility index.
+
+        Returns
+        -------
+        float
+            The sum of coefficients in absolute value.
+
         """
-            Calculate the sum of all coefficients (excluding water) in
-            absolute value.
-            This is useful for calculating the reversibility index.
-        """
+
         return sum(abs(c) for _, c in self.items(protons=False, water=False))
 
-    def check_half_reaction_balancing(self):
-        """
-            :return: The number of electrons that are 'missing' in the
+    def check_half_reaction_balancing(self) -> Union[int, None]:
+        """Check if this is a balanced half-reaction (redox).
+
+        :return:
+
+        Returns
+        -------
+        int or None
+            The number of electrons that are 'missing' in the
             half-reaction or None if the reaction is not atomwise-balanced.
         """
+
         atom_bag = self._get_reaction_atom_bag()
         if atom_bag is None:
             return None
@@ -485,12 +677,21 @@ class Reaction:
         If it is equal to another reaction's string, then they have identical
         stoichiometry.
 
-        :param sparse: a dictionary whose keys are compounds and values are
-        stoichiometric coefficients
-        :return: a unique list of pairs (compound.id, coefficient)
+        Parameters
+        ----------
+        sparse : Dict[Compound, float]
+            a dictionary whose keys are compounds and values are
+            stoichiometric coefficients
+
+
+        Returns
+        -------
+        Tuple[Tuple[int, float]]
+            a tuple containing all pairs of (compound.id, coefficient)
+
         """
         if len(sparse) == 0:
-            return []
+            return tuple()
 
         # sort according to compound ID and normalize the stoichiometric
         # coefficients such that the coeff of the reactant with the lowest
@@ -505,7 +706,8 @@ class Reaction:
             filter(lambda x: x[1] != 0, sorted_cpd_coeff_list)
         )
         if len(sorted_cpd_coeff_list) == 0:
-            raise Exception("All stoichiometric ceofficients are 0")
+            warnings.warn("All stoichiometric coefficients are 0")
+            return tuple()
         norm_factor = 1.0 / sorted_cpd_coeff_list[0][1]
         r_list = [
             (cpd.id, norm_factor * coeff)
@@ -520,7 +722,11 @@ class Reaction:
         stoichiometry. We create a unique formula string based on the
         Compound IDs and coefficients.
 
-        :return: a hash of the Reaction.
+        Returns
+        -------
+        int
+            A hash of the Reaction.
+
         """
         reactants = Reaction._hashable_reactants(self.sparse)
         return hash(reactants)
@@ -532,26 +738,29 @@ def create_stoichiometric_matrix_from_reactions(
     is_water: Callable[[Compound], bool],
     water: Compound,
 ) -> pd.DataFrame:
-    """Builds a stoichiometric matrix.
+    """Build a stoichiometric matrix.
 
     Parameters
     ----------
-    reactions : iterable
-        The collection of reactions to build a stoichiometric matrix from.
+    reactions : Iterable[Reaction]
+        The collection of reactions to build a stoichiometric
+        matrix from.
     is_proton : Callable[[Compound], bool]
         A function that identifies if a Compound is a proton.
     is_water : Callable[[Compound], bool]
-         A function that identifies if a Compound is water.
+        A function that identifies if a Compound is water.
     water : Compound
-        The water compound (is added to the stoichiometric matrix if water is
-        not one of the compounds in the input reaction list.
+        The water compound (is added to the stoichiometric matrix
+        if water is not one of the compounds in the input reaction list.
+
 
     Returns
     -------
-    pandas.DataFrame
+    DataFrame
         The stoichiometric matrix as a DataFrame whose indexes are the
         compounds and its columns are the reactions (in the same order as
         the input).
+
     """
     # Now get rid of the protons, since we are applying Alberty's
     # framework where their potential is set to 0, and the pH is held
